@@ -382,7 +382,7 @@ let reset_line_indent config current_line path =
             when acc1.line_indent = acc1.column
             ->
               (* ignore those if at start of line *)
-              acc1 :: p, acc, if acc1.kind = KBracketBar then 2 else 1
+              acc1 :: p, acc, acc1.pad
           | _ -> p, acc, 0
         in
         List.fold_left (fun p t ->
@@ -428,7 +428,6 @@ let op_prio_align_indent config =
   (* anything else : -10 *)
   (* in -> : 0 *)
   | SEMI -> prio_semi,L,-2
-  | AS -> 8,L,0
   (* special negative indent is only honored at beginning of line *)
   (* then else : 10 *)
   | BAR -> 10,T,-2
@@ -436,7 +435,8 @@ let op_prio_align_indent config =
   | LESSMINUS | COLONEQUAL -> 20,L,config.i_base
   | COMMA -> 30,align,-2
   | MINUSGREATER -> 32,L,0 (* is an operator only in types *)
-  | COLON | COLONGREATER -> 35,L,config.i_base
+  | COLON -> 35,T,config.i_base
+  | COLONGREATER -> 35,L,config.i_base
   | OR | BARBAR -> 40,T,0
   | AMPERSAND | AMPERAMPER -> 50,T,0
   | (INFIXOP0 s | INFIXOP1 s | INFIXOP2 s | INFIXOP3 s | INFIXOP4 s)
@@ -454,7 +454,8 @@ let op_prio_align_indent config =
   | INFIXOP3 _ | STAR -> 100,align,indent
   | INFIXOP4 _ -> 110,align,indent
   (* apply: 140 *)
-  | TILDE | QUESTION -> 140,L,config.i_base
+  | AS -> prio_apply,L,0
+  | TILDE | QUESTION -> prio_apply,L,config.i_base
   | LABEL _ | OPTLABEL _ ->
       if config.i_align_params = Always then 145,T,config.i_base
       else 145,L,config.i_base
@@ -462,12 +463,60 @@ let op_prio_align_indent config =
   | DOT -> prio_dot,align,config.i_base
   | _ -> assert false
 
+let handle_dotted block tok =
+  let starts_line = tok.newlines > 0 in
+  let current_line = Region.start_line tok.region in
+  let is_attr_id = function
+    | { kind = KAttrId (_, dotted) } :: _ -> not dotted
+    | _ -> false in
+  let make_dotted_attr_id = function
+    | { kind = KAttrId (names, _) } as node ::
+      ({ kind = (KExtendedItem [] | KExtendedExpr [])} :: _ as path) ->
+        { node with kind = KAttrId (names, true) } :: path
+    | _ -> assert false in
+  let is_dotted_attr_id = function
+    | { kind = KExtendedExpr [] } :: _ -> true
+    | { kind = KExtendedItem [] } :: _ -> true
+    | { kind = KAttrId (_, dotted) } :: _ -> dotted
+    | _ -> false in
+  let make_attr_id name = function
+    | ({ kind = (KExtendedItem [] | KExtendedExpr []);
+         indent; pad; } :: _ as path) ->
+          let indent =
+            if starts_line then indent + pad
+            else indent + pad + String.length (Lazy.force tok.between) - 1 in
+          let column =
+            if starts_line then indent else block.toff + tok.offset in
+          { kind = (KAttrId ([name], false)); indent;
+            line_indent = indent; column; line = current_line;
+            pad = 0 } :: path
+    | ({ kind = KAttrId (names, _)} as node) :: path ->
+        { node with kind = KAttrId (name :: names, false); } :: path
+    | _ -> assert false in
+  if is_dotted_attr_id block.path then
+    match tok.token with
+    | LIDENT s | UIDENT s ->
+        Some (make_attr_id s block.path)
+    | AND | AS | ASSERT | BEGIN | CLASS | CONSTRAINT | DO | DONE
+    | DOWNTO | ELSE | END | EXCEPTION | EXTERNAL | FALSE | FOR | FUN
+    | FUNCTION | FUNCTOR | IF | IN | INCLUDE | INHERIT | INITIALIZER
+    | LAZY | LET | MATCH | METHOD | MODULE | MUTABLE | NEW | OBJECT | OF
+    | OPEN | OR | PRIVATE | REC | SIG | STRUCT | THEN | TO | TRUE | TRY
+    | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
+        Some (make_attr_id (Lazy.force tok.substr) block.path)
+    | _ -> None
+  else if is_attr_id block.path then
+    match tok.token with
+    | DOT -> Some (make_dotted_attr_id block.path)
+    | _ -> None
+  else
+    None
+
 (* Take a block, a token stream and a token.
    Return the new block stack. *)
 let rec update_path config block stream tok =
   let open IndentConfig in
-  let is_first_token = Region.start_line tok.region = 0 in
-  let starts_line = tok.newlines > 0 || is_first_token in
+  let starts_line = tok.newlines > 0 in
   let current_line = Region.start_line tok.region in
   let node replace kind pos pad path =
     let parent = Path.top path in
@@ -496,7 +545,7 @@ let rec update_path config block stream tok =
         let path =
           let indent =
             if starts_line then h2.indent else block.toff + tok.offset in
-          let pad = if starts_line then 2 else 0 in
+          let pad = if starts_line then config.i_base else 0 in
           {h1 with indent; column=indent; pad } :: path in
         node false kind pos pad path :: path
     | {kind = KAttrId (names, _)} ::
@@ -540,8 +589,8 @@ let rec update_path config block stream tok =
                   paren.column + paren_len + 1 (* usually 1 space *) + pad
                 in
                 Some ({ h with kind; indent; column=indent;
-                               line_indent = indent;
-                               pad = max h.pad (h.indent-indent) } :: p)
+                               line_indent = indent-pad;
+                               pad = max h.pad (h.indent-indent)} :: p)
             | _ ->
                 match kind,h.kind with
                 | KExpr pk, KExpr ph when ph = pk ->
@@ -587,6 +636,10 @@ let rec update_path config block stream tok =
           match unwind_while (fun kind -> prio kind >= prio_apply) path with
           | Some({kind=KExpr i} as e1 :: p) when i = prio_apply ->
               {e1 with line_indent = e.line_indent} :: p
+          | Some({kind=KExpr _; line} ::
+                 {kind=KModule|KInclude|KOpen|KBody KModule} :: _
+                 as p) -> (* ignore align_params for functor application *)
+              extend (KExpr prio_apply) L (reset_line_indent config line p)
           | Some({kind=KExpr _; line}
               :: {kind=KArrow (KMatch|KTry) | KTry | KMatch;
                   line=arrow_line}::_ as p)
@@ -698,46 +751,9 @@ let rec update_path config block stream tok =
     | {kind=KComment _|KVerbatim _|KUnknown}::path -> {block with path}
     | _ -> block
   in
-  let is_attr_id = function
-    | { kind = KAttrId (_, dotted) } :: _ -> not dotted
-    | _ -> false in
-  let make_dotted_attr_id = function
-    | { kind = KAttrId (names, _) } as node ::
-      ({ kind = (KExtendedItem [] | KExtendedExpr [])} :: _ as path) ->
-        { node with kind = KAttrId (names, true) } :: path
-    | _ -> assert false in
-  let is_dotted_attr_id = function
-    | { kind = KExtendedExpr [] } :: _ -> true
-    | { kind = KExtendedItem [] } :: _ -> true
-    | { kind = KAttrId (_, dotted) } :: _ -> dotted
-    | _ -> false in
-  let make_attr_id name = function
-    | ({ kind = (KExtendedItem [] | KExtendedExpr []);
-         indent; pad; } :: _ as path) ->
-          let indent =
-            if starts_line then indent + pad
-            else indent + pad + String.length (Lazy.force tok.between) - 1 in
-          let column =
-            if starts_line then indent else block.toff + tok.offset in
-          { kind = (KAttrId ([name], false)); indent;
-            line_indent = indent; column; line = current_line;
-            pad = 0 } :: path
-    | ({ kind = KAttrId (names, _)} as node) :: path ->
-        { node with kind = KAttrId (name :: names, false); } :: path
-    | _ -> assert false in
+  let (>>!) opt f = match opt with Some x -> x | None -> f () in
+  handle_dotted block tok >>! fun () ->
   match tok.token with
-  | LIDENT s | UIDENT s when is_dotted_attr_id block.path ->
-      make_attr_id s block.path
-  | AND | AS | ASSERT | BEGIN | CLASS | CONSTRAINT | DO | DONE
-  | DOWNTO | ELSE | END | EXCEPTION | EXTERNAL | FALSE | FOR | FUN
-  | FUNCTION | FUNCTOR | IF | IN | INCLUDE | INHERIT | INITIALIZER
-  | LAZY | LET | MATCH | METHOD | MODULE | MUTABLE | NEW | OBJECT | OF
-  | OPEN | OR | PRIVATE | REC | SIG | STRUCT | THEN | TO | TRUE | TRY
-  | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH
-    when is_dotted_attr_id block.path ->
-      make_attr_id (Lazy.force tok.substr) block.path
-  | DOT when is_attr_id block.path ->
-      make_dotted_attr_id block.path
   | SEMISEMI    -> append KUnknown L ~pad:0 (unwind_top block.path)
   | INCLUDE     -> append KInclude L (unwind_top block.path)
   | EXCEPTION   ->
@@ -799,6 +815,7 @@ let rec update_path config block stream tok =
       (match fold_expr block.path with
        | l :: _ as p
          when not starts_line
+           && l.kind <> KExpr 0
            && (config.i_strict_with = Never
                || config.i_strict_with = Auto && l.kind <> KBegin) ->
            let p = reset_line_indent config current_line p in
@@ -818,20 +835,31 @@ let rec update_path config block stream tok =
                 replace KFun L path
             | _ -> append KFun L block.path)
        | p -> append KFun L (fold_expr p))
-  | STRUCT ->
-      let path =
-        reset_line_indent config current_line block.path
+  | STRUCT | SIG ->
+      let k = match tok.token with
+        | STRUCT -> KStruct
+        | SIG -> KSig
+        | _ -> assert false
       in
-      append KStruct L (reset_padding path)
+      let expr_start =
+        unwind (function KParen | KLet | KLetIn | KBody _ -> true | _ -> false)
+          block.path
+      in
+      let indent = match expr_start with
+        | {kind=KParen}::{kind=KExpr prio; line; indent}::_
+          when prio = prio_apply && line = current_line ->
+            indent
+        | _ -> Path.indent block.path
+      in
+      Path.maptop (fun n -> {n with indent})
+        (append k L (reset_padding block.path))
+
   | WHEN ->
       append KWhen L ~pad:(config.i_base + if starts_line then 0 else 2)
         (unwind (function
            | KWith(KTry|KMatch) | KBar(KTry|KMatch) | KFun -> true
            | _ -> false)
            block.path)
-  | SIG ->
-      append KSig L (reset_padding block.path)
-
   | OPEN ->
       if last_token block = Some LET then
         append KOpen L block.path
@@ -944,8 +972,10 @@ let rec update_path config block stream tok =
            let path = unwind (function
                |KTry|KMatch
                |KVal|KType|KBody KType|KException (* type-conv *)
-               |KWith KTry (* useful for lwt's try-finally *)
+               |KColon
                |KBrace -> true
+               |KWith KTry -> (* useful for lwt's try-finally *)
+                   Lazy.force tok.substr = "finally"
                | _ -> false
              ) block.path in
            match path with
@@ -980,6 +1010,9 @@ let rec update_path config block stream tok =
                else
                  replace (KWith kind) L ~pad:config.i_with
                    (reset_line_indent config current_line path)
+           | {kind=KColon}::_ as p ->
+               (* may happen with sexp extension, 'with default' *)
+               append expr_atom L p
            | _ -> path)
 
   | IF ->
@@ -1071,9 +1104,9 @@ let rec update_path config block stream tok =
             path
         in
         match path with
-        | {kind=KFun} :: {kind=KExpr i} :: path when i = prio_flatop ->
+        | {kind=KFun} :: ({kind=KExpr i} as e) :: path when i = prio_flatop ->
             (* eg '>>= fun x ->': indent like the top of the expression *)
-            path
+            {e with kind = KExpr 0} :: path
         | {kind=KFun; line } :: _
           when next_offset tok stream = None
             && line = current_line
@@ -1154,7 +1187,8 @@ let rec update_path config block stream tok =
                let h = {h with indent = h.indent + indent; pad = 0} in
                replace (KBody kind) L ~pad:0 (h :: p)
              else
-               replace (KBody kind) L ~pad:indent (h :: p))
+               let h = {h with indent = h.column} in
+               replace (KBody kind) T ~pad:indent (h :: p))
       in
       find_parent block.path
 
@@ -1180,7 +1214,7 @@ let rec update_path config block stream tok =
          | KAnd(KModule|KLet|KLetIn|KExternal)} :: _ ->
            append KColon L path
        | {kind=KVal} :: {kind=KObject} :: _ ->
-           make_infix tok block.path
+           make_infix tok path
        | {kind=KVal} as h :: p ->
            let indent = config.i_base in
            if starts_line then
@@ -1235,15 +1269,6 @@ let rec update_path config block stream tok =
            { h with pad = config.i_base } :: p
        | _ -> make_infix tok block.path)
 
-  | AMPERAMPER | BARBAR ->
-      (* back-indented when after if or when and not alone *)
-      let op_prio, _align, _indent = op_prio_align_indent config tok.token in
-      (match unwind_while (fun kind -> prio kind >= op_prio) block.path with
-       | Some ({kind=KExpr _; line}::{kind=KWhen|KIf; line=line_if}::_ as p)
-         when line = line_if && next_offset tok stream <> None ->
-           extend (KExpr op_prio) T ~pad:(-3) p
-       | _ -> make_infix tok block.path)
-
   | LESS ->
       if is_inside_type block.path then
         (* object type *)
@@ -1267,6 +1292,7 @@ let rec update_path config block stream tok =
 
   | LESSMINUS | COMMA | OR
   | AMPERSAND | INFIXOP0 _ | INFIXOP1 _
+  | AMPERAMPER | BARBAR
   | COLONCOLON | INFIXOP2 _ | PLUSDOT | PLUS | MINUSDOT | MINUS
   | INFIXOP3 _ | STAR | INFIXOP4 _
   | SHARP | AS | COLONGREATER
@@ -1288,7 +1314,8 @@ let rec update_path config block stream tok =
 
   | UIDENT _ ->
       (match block.path with
-       | {kind=KBody KType}::_ when starts_line ->
+       | {kind=KBody KType}::_
+         when starts_line && next_token stream <> Some DOT ->
            (* type =\nA\n| B : append a virtual bar before A for alignment *)
            let path = append (KBar KType) L ~pad:2 block.path
            in atom path
@@ -1301,13 +1328,20 @@ let rec update_path config block stream tok =
            in
            let path = append (KBar KType) L ~pad:2 path
            in atom path
+       | {kind=KModule | KInclude | KOpen}::_ when not starts_line ->
+           (* indent functor parameters as if indent was flushed (like after a
+              newline) *)
+           Path.maptop (fun n ->
+               let indent = n.indent + n.pad in
+               {n with indent; line_indent = indent; pad = config.i_base}
+             ) (atom block.path)
        | _ -> atom block.path)
 
   | INT64 _ | INT32 _ | INT _ | LIDENT _
-  | FLOAT _ | CHAR _ | STRING _ | EOF_IN_STRING _
+  | FLOAT _ | CHAR _ | STRING _
   | TRUE | FALSE | NATIVEINT _
   | UNDERSCORE | TILDE | QUESTION
-  | QUOTE | QUOTATION | EOF_IN_QUOTATION _ ->
+  | QUOTE | QUOTATION ->
       atom block.path
 
   | PREFIXOP _ | BANG | QUESTIONQUESTION ->
@@ -1351,7 +1385,7 @@ let rec update_path config block stream tok =
        | {kind=KCodeInComment|KVerbatim _} :: p -> p
        | _ -> block.path)
 
-  | COMMENT | EOF_IN_COMMENT ->
+  | COMMENT ->
       let s = Lazy.force tok.substr in
       let pad =
         let len = String.length s in
@@ -1411,10 +1445,7 @@ let rec update_path config block stream tok =
                      append (KComment (tok,indent)) (A indent) ~pad block.path
                  | None ->  (* recursive call to indent like next line *)
                      let path = match next with
-                       | Some ({token = EOF | EOF_IN_COMMENT |
-                                EOF_IN_STRING _ | EOF_IN_QUOTATION _}
-                              , _)
-                       | None -> []
+                       | Some ({token = EOF }, _) | None -> []
                        | Some (next,stream) ->
                            update_path config block stream next
                      in
@@ -1440,7 +1471,7 @@ let update config block stream tok =
   let path = update_path config block stream tok in
   let last = match tok.token with
     | COMMENT | COMMENTCONT | OCAMLDOC_VERB
-    | EOF | EOF_IN_COMMENT | EOF_IN_QUOTATION _ | EOF_IN_STRING _ ->
+    | EOF ->
         tok :: block.last
     | _ -> [tok] in
   let toff =
@@ -1509,7 +1540,7 @@ let guess_indent line t =
     -> (* Inside comment *)
       Path.indent t.path + Path.pad t.path
   | {kind=KExpr i}::p,
-    ({token=EOF|EOF_IN_COMMENT|EOF_IN_QUOTATION _|EOF_IN_STRING _} :: tok :: _
+    ({token=EOF} :: tok :: _
     | tok::_)
     when i = prio_max
       && line > Region.end_line tok.region + 1
