@@ -48,7 +48,7 @@ ocp-indent configuration files"
 (defcustom ocp-indent-syntax nil
   "*Enabled syntax extensions for ocp-indent (see option --syntax)"
   :group 'ocp-indent
-  :type '(repeat (string)))
+  :type '(repeat string))
 
 (defcustom ocp-indent-allow-tabs nil
   "*Allow indent-tabs-mode in ocaml buffers. Not recommended, won't work well."
@@ -81,7 +81,8 @@ are blanks."
   (let*
       ((start-line (line-number-at-pos start))
        (end-line (line-number-at-pos end))
-       (errfile (make-temp-name "ocp-indent-error"))
+       (errfile (expand-file-name (make-temp-name "ocp-indent-error")
+                                  temporary-file-directory))
        (indents-str
         (with-output-to-string
           (if (/= 0
@@ -105,6 +106,10 @@ are blanks."
   (interactive nil)
   (ocp-indent-region (point) (point)))
 
+(defun ocp-indent-buffer ()
+  (interactive nil)
+  (ocp-indent-region 0 (buffer-size)))
+
 ;;;###autoload
 (defun ocp-setup-indent ()
   (interactive nil)
